###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import pytest

from pidcalib2 import binning


def test_p_binning():
    with pytest.raises(KeyError):
        binning.p_binning("graviton")

    reference_K_pi_p_binning = [
        3000,
        9300,
        15600,
        19000.0,
        24400.0,
        29800.0,
        35200.0,
        40600.0,
        46000.0,
        51400.0,
        56800.0,
        62200.0,
        67600.0,
        73000.0,
        78400.0,
        83800.0,
        89200.0,
        94600.0,
        100000.0,
    ]
    assert binning.p_binning("K") == reference_K_pi_p_binning
    assert binning.p_binning("Pi") == reference_K_pi_p_binning
    assert binning.p_binning("P") == reference_K_pi_p_binning

    assert binning.p_binning("Mu") == [
        3000,
        6000,
        8000,
        10000,
        12000,
        14500,
        17500,
        21500,
        27000,
        32000,
        40000,
        60000,
        70000,
        100000.0,
    ]

    with pytest.raises(KeyError):
        binning.p_binning("pi")

    with pytest.raises(KeyError):
        binning.p_binning("Nu")


def test_set_binning():
    with pytest.raises(TypeError):
        binning.set_binning("Pi", "P", 30)  # type: ignore

    binning.set_binning("GhostParticle", "P", [10, 20])
    assert binning.BINNINGS["GhostParticle"]["P"] == {"bin_edges": [10, 20]}


def test_get_binning():
    with pytest.raises(KeyError):
        binning.get_binning("PION", "P")

    with pytest.raises(KeyError):
        binning.get_binning("Pi", "Non-existing var")



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import math
import os
from pathlib import Path

import uproot
import pytest

from pidcalib2 import pid_data


@pytest.fixture
def test_path():
    return Path(os.path.dirname(os.path.abspath(__file__)))


@pytest.mark.xrootd
@pytest.mark.slow
def test_root_to_dataframe():
    assert (
        pid_data.root_to_dataframe(
            "root://eoslhcb.cern.ch//eos/lhcb/grid/prod/lhcb/LHCb/Collision15/PIDCALIB.ROOT/00064787/0000/00064787_00000037_1.pidcalib.root",  # noqa: E501
            ["DSt_PiMTuple/DecayTree", "DSt_PiPTuple/DecayTree"],
            ["probe_PIDK"],
        ).shape[0]
        == 18400
    )

    with pytest.raises(uproot.exceptions.KeyInFileError):
        pid_data.root_to_dataframe(
            "root://eoslhcb.cern.ch//eos/lhcb/grid/prod/lhcb/LHCb/Collision15/PIDCALIB.ROOT/00064787/0000/00064787_00000037_1.pidcalib.root",  # noqa: E501
            ["DSt_PiMTuple/DecayTree", "DSt_PiPTuple/DecayTree"],
            ["probe_PIDnu"],
        )

    with pytest.raises(uproot.exceptions.KeyInFileError):
        pid_data.root_to_dataframe(
            "root://eoslhcb.cern.ch//eos/lhcb/grid/prod/lhcb/LHCb/Collision15/PIDCALIB.ROOT/00064787/0000/00064787_00000037_1.pidcalib.root",  # noqa: E501
            ["DSt_PiMTuple/DecayTree", "DSt_PiPTuple/DecayTree"],
            ["DLLnu"],
            True,
        )


def test_get_tree_paths():
    assert pid_data.get_tree_paths("Pi", "26") == ["DecayTree"]
    assert pid_data.get_tree_paths("Pi", "Turbo15") == [
        "DSt_PiMTuple/DecayTree",
        "DSt_PiPTuple/DecayTree",
    ]
    assert pid_data.get_tree_paths("Mu", "Turbo15-MagUp-Mu") == [
        "Jpsi_MuPTuple/DecayTree",
        "Jpsi_MuMTuple/DecayTree",
    ]
    assert pid_data.get_tree_paths("Mu", "Turbo15-MagUp-Mu", ["Tree1", "Tree2"]) == [
        "Tree1/DecayTree",
        "Tree2/DecayTree",
    ]


def test_get_relevant_branch_names():
    assert pid_data.get_relevant_branch_names(["DLLK < 4"], ["P"]) == {
        "sWeight": "probe_sWeight",
        "P": "probe_P",
        "DLLK": "probe_PIDK",
    }

    assert pid_data.get_relevant_branch_names(["DLLp > 4"], ["P", "ETA"]) == {
        "sWeight": "probe_sWeight",
        "P": "probe_P",
        "ETA": "probe_ETA",
        "DLLp": "probe_PIDp",
    }

    assert pid_data.get_relevant_branch_names(["DLLp == 4"], ["P", "ETA"]) == {
        "sWeight": "probe_sWeight",
        "P": "probe_P",
        "ETA": "probe_ETA",
        "DLLp": "probe_PIDp",
    }

    assert pid_data.get_relevant_branch_names(["DLLp != 4"], ["P", "ETA"]) == {
        "sWeight": "probe_sWeight",
        "P": "probe_P",
        "ETA": "probe_ETA",
        "DLLp": "probe_PIDp",
    }

    assert pid_data.get_relevant_branch_names(["test_DLL != 4"], ["special_var"]) == {
        "sWeight": "probe_sWeight",
        "test_DLL": "test_DLL",
        "special_var": "special_var",
    }

    with pytest.raises(KeyError):
        pid_data.get_relevant_branch_names(["DLLp > 4 & probe_PIDp > 4"], ["P", "ETA"])

    assert pid_data.get_relevant_branch_names(
        ["test_DLL != 4"], ["special_var"], ["raw_var > 0"]
    ) == {
        "sWeight": "probe_sWeight",
        "test_DLL": "test_DLL",
        "special_var": "special_var",
        "raw_var": "raw_var",
    }

    assert pid_data.get_relevant_branch_names(
        ["test_DLL != 4"], ["special_var"], ["P > 1000"]
    ) == {
        "sWeight": "probe_sWeight",
        "test_DLL": "test_DLL",
        "special_var": "special_var",
        "P": "probe_P",
    }


def test_dataframe_from_local_file(test_path):
    df = pid_data.dataframe_from_local_file(
        str(test_path / "test_data/cal_test_data.csv"), ["sWeight"]
    )
    assert df.shape[0] == 99
    assert df["sWeight"][0] == pytest.approx(1.1081801082842266)

    df_pkl = pid_data.dataframe_from_local_file(
        str(test_path / "test_data/cal_test_data.pkl"), ["sw"]
    )
    assert df_pkl.shape[0] == 99
    assert df_pkl["sw"][0] == pytest.approx(1.1081801082842266)

    with pytest.raises(KeyError):
        pid_data.dataframe_from_local_file(
            str(test_path / "test_data/cal_test_data.csv"),
            ["this key doesn't exist"],
        )

    with pytest.raises(Exception):
        pid_data.dataframe_from_local_file(
            str(test_path / "test_data/file.root"), ["sWeight"]
        )


def test_get_reference_branch_names():
    ref_pars = {"Bach": ["K", "DLLK > 4"]}
    bin_vars = {"P": "P", "ETA": "ETA", "nTracks": "nTracks"}
    branch_names = pid_data.get_reference_branch_names(ref_pars, bin_vars)
    assert branch_names == ["Bach_P", "Bach_ETA", "nTracks"]


def test_get_reference_branch_name():
    assert pid_data.get_reference_branch_name("Bach", "foo", "bar") == "Bach_bar"
    assert pid_data.get_reference_branch_name("Bach", "nTracks", "bar") == "bar"
    assert pid_data.get_reference_branch_name("Bach", "nSPDhits", "bar") == "bar"
    assert (
        pid_data.get_reference_branch_name("Bach", "foo", "nTracks") == "Bach_nTracks"
    )


def test_get_calib_hists(test_path):
    ref_pars = {"Bach": ["K", "DLLK > 4"]}
    bin_vars = {"P": "P", "ETA": "ETA", "nTracks": "nTracks"}
    eff_hists = pid_data.get_calib_hists(
        str(test_path / "test_data"), "Turbo18", "up", ref_pars, bin_vars
    )
    assert math.isnan(eff_hists["Bach"]["eff"].sum().value)  # type: ignore
    assert eff_hists["Bach"]["eff"][4, 2, 1].value == pytest.approx(  # type: ignore
        0.9759721725381351
    )
    assert eff_hists["Bach"]["passing"].sum().value == pytest.approx(  # type: ignore
        77882.51311058077
    )
    assert eff_hists["Bach"]["total"].sum().value == pytest.approx(  # type: ignore
        86103.98549868543
    )
    assert eff_hists["Bach"]["passing"].sum().variance == pytest.approx(  # type: ignore
        97895.25547125406
    )
    assert eff_hists["Bach"]["total"].sum().variance == pytest.approx(  # type: ignore
        111911.08518551107
    )

    with pytest.raises(FileNotFoundError):
        pid_data.get_calib_hists(
            str(test_path / "test_data"),
            "Turbo34",
            "up",
            ref_pars,
            bin_vars,
        )


def test_get_calibration_sample(test_path):
    assert (
        len(
            pid_data.get_calibration_sample(
                "Turbo18", "up", "Pi", str(test_path / "../data/samples.json")
            )["files"]
        )
        == 428
    )
    assert (
        len(
            pid_data.get_calibration_sample(
                "Turbo18",
                "up",
                "Pi",
                str(test_path / "../data/samples.json"),
                3,
            )["files"]
        )
        == 3
    )
    assert (
        len(
            pid_data.get_calibration_sample(
                "Electron18",
                "down",
                "e_B_Jpsi",
                str(test_path / "../data/samples.json"),
            )["files"]
        )
        == 1
    )
    assert (
        len(
            pid_data.get_calibration_sample(
                "20", "down", "K", str(test_path / "../data/samples.json")
            )["files"]
        )
        == 72
    )

    with pytest.raises(KeyError):
        pid_data.get_calibration_sample(
            "Turbo34", "up", "Pi", str(test_path / "../data/samples.json")
        )



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import copy
import math
import os
from pathlib import Path

import pandas as pd
import pytest

from pidcalib2 import binning, make_eff_hists


@pytest.fixture
def test_path():
    return Path(os.path.dirname(os.path.abspath(__file__)))


@pytest.fixture
def config():
    return {
        "bin_vars": None,
        "binning_file": None,
        "cuts": None,
        "file_list": None,
        "list": None,
        "local_dataframe": None,
        "magnet": None,
        "output_dir": None,
        "particle": None,
        "pid_cuts": None,
        "sample": None,
        "samples_file": None,
        "verbose": True,
    }


def test_make_eff_hists(config, test_path, tmp_path):
    config.update(
        {
            "bin_vars": ["P"],
            "local_dataframe": str(test_path / "test_data/cal_test_data.csv"),
            "magnet": "up",
            "output_dir": str(tmp_path / "test_output"),
            "particle": "Pi",
            "pid_cuts": ["DLLK < 4", "DLLK<3"],
            "sample": "Turbo18",
        }
    )
    make_eff_hists.make_eff_hists(config)
    eff_histo = pd.read_pickle(
        tmp_path / "test_output/effhists-Turbo18-up-Pi-DLLK<4-P.pkl"
    )

    # These asserts might come in handy when some detail in the boost_histogram
    # implementation changes, thus failing the reference histogram comparison.
    assert eff_histo.sum(flow=False).value == pytest.approx(17.812484216412948)
    assert eff_histo.sum(flow=True).value == pytest.approx(23.57053953265155)
    assert eff_histo[1].value == 0.8978157343833105

    eff_histo_reference = pd.read_pickle(
        test_path / "test_data/effhists-Turbo18-up-pi-DLLK<4-P.pkl"
    )
    assert eff_histo == eff_histo_reference


def test_make_eff_hists_with_user_cuts(config, test_path, tmp_path):
    config.update(
        {
            "bin_vars": ["P"],
            "cuts": ["Dst_IPCHI2 < 100", "probe_TRACK_GHOSTPROB < 0.05"],
            "local_dataframe": str(test_path / "test_data/cal_test_data.csv"),
            "magnet": "up",
            "output_dir": str(tmp_path / "test_output"),
            "particle": "Pi",
            "pid_cuts": ["DLLK < 4", "DLLK<3"],
            "sample": "Turbo18",
        }
    )
    make_eff_hists.make_eff_hists(config)
    eff_histo = pd.read_pickle(
        tmp_path / "test_output/effhists-Turbo18-up-Pi-DLLK<4-P.pkl"
    )

    # These asserts might come in handy when some detail in the boost_histogram
    # implementation changes, thus failing the reference histogram comparison.
    assert eff_histo.sum(flow=False).value == pytest.approx(17.812484216412948)
    assert eff_histo.sum(flow=True).value == pytest.approx(23.57053953265155)
    assert eff_histo[1].value == 0.8978157343833105

    eff_histo_reference = pd.read_pickle(
        test_path / "test_data/effhists-Turbo18-up-pi-DLLK<4-P.pkl"
    )
    assert eff_histo == eff_histo_reference

    # Test that stricter cuts have an effect
    config["cuts"] = ["Dst_IPCHI2 < 9", "probe_TRACK_GHOSTPROB < 0.01"]
    make_eff_hists.make_eff_hists(config)
    eff_histo = pd.read_pickle(
        tmp_path / "test_output/effhists-Turbo18-up-Pi-DLLK<4-P.pkl"
    )

    # With stricter cuts, some bins in the efficiency histogram become empty
    assert math.isnan(eff_histo.sum(flow=False).value)
    assert math.isnan(eff_histo.sum(flow=True).value)
    assert eff_histo[1].value == 0.806456926487749

    assert eff_histo != eff_histo_reference


@pytest.mark.xrootd
@pytest.mark.slow
def test_make_eff_hists_with_hardcoded_cuts(config, tmp_path):
    config.update(
        {
            "bin_vars": ["P"],
            "magnet": "down",
            "max_files": 1,
            "output_dir": str(tmp_path / "test_output"),
            "particle": "P_IncLc",
            "pid_cuts": ["DLLp > 4"],
            "sample": "Turbo16",
        }
    )
    make_eff_hists.make_eff_hists(config)
    eff_histo = pd.read_pickle(
        tmp_path / "test_output/effhists-Turbo16-down-P_IncLc-DLLp>4-P.pkl"
    )

    # These asserts might come in handy when some detail in the boost_histogram
    # implementation changes, thus failing the reference histogram comparison.
    assert eff_histo.sum(flow=False) == eff_histo.sum(flow=True)
    assert eff_histo.sum(flow=True).value == pytest.approx(17.3387622437092)
    assert eff_histo[1].value == 0.9686555826741122


@pytest.mark.xrootd
@pytest.mark.slow
def test_make_eff_hists_local_file_list(config, test_path, tmp_path):
    config.update(
        {
            "bin_vars": ["P"],
            "file_list": str(test_path / "test_data/test_file_list"),
            "magnet": "up",
            "output_dir": str(tmp_path / "test_output"),
            "particle": "Pi",
            "pid_cuts": ["DLLK < 4", "DLLK<3"],
            "sample": "Turbo18",
        }
    )
    make_eff_hists.make_eff_hists(config)
    eff_histo = pd.read_pickle(
        tmp_path / "test_output/effhists-Turbo18-up-Pi-DLLK<4-P.pkl"
    )

    # These asserts might come in handy when some detail in the boost_histogram
    # implementation changes, thus failing the reference histogram comparison.
    assert eff_histo.sum(flow=False).value == pytest.approx(17.378390835392615)
    assert eff_histo[1].value == 0.9797018332573063


@pytest.mark.xrootd
@pytest.mark.slow
def test_make_eff_hists_file_list(config, test_path, tmp_path):
    config.update(
        {
            "bin_vars": ["P"],
            "magnet": "up",
            "max_files": 1,
            "output_dir": str(tmp_path / "test_output"),
            "particle": "Pi",
            "pid_cuts": ["DLLK < 4", "DLLK<3"],
            "sample": "Turbo18",
            "samples_file": str(test_path / "test_data/test_samples.json"),
        }
    )
    make_eff_hists.make_eff_hists(config)
    eff_histo = pd.read_pickle(
        tmp_path / "test_output/effhists-Turbo18-up-Pi-DLLK<4-P.pkl"
    )

    # These asserts might come in handy when some detail in the boost_histogram
    # implementation changes, thus failing the reference histogram comparison.
    assert eff_histo.sum(flow=False).value == pytest.approx(17.378390835392615)
    assert eff_histo[1].value == 0.9797018332573063


def test_decode_arguments(capsys, test_path, tmp_path):
    assert (
        make_eff_hists.decode_arguments(
            [
                "--sample=Turbo18",
                "-m=up",
                "--particle=K",
                "--pid-cut='DLLK>5'",
                "--bin-var=P",
            ]
        ).verbose
        is False
    )

    with pytest.raises(SystemExit):
        make_eff_hists.decode_arguments(
            ["-m=up", "--particle=K", "--pid-cut='DLLK>5'", "--bin-var=P"]
        )

    with pytest.raises(SystemExit):
        make_eff_hists.decode_arguments(
            ["--sample=Turbo18", "-m=up", "--particle=K", "--bin-var=P"]
        )

    with pytest.raises(SystemExit):
        make_eff_hists.decode_arguments(
            ["--list", str(test_path / "test_data/test_samples.json")]
        )
    captured = capsys.readouterr().out
    assert captured == (
        "Sample  | Magnet | Particle\n"
        "--------|--------|---------\n"
        "Turbo18 | up     | Pi\n"
    )

    with pytest.raises(SystemExit):
        make_eff_hists.decode_arguments(
            ["--list", "aliases"]
        )
    captured = capsys.readouterr().out
    assert captured.startswith(
        "Alias                         | Variable\n"
        "------------------------------|------------------------------------\n"
    )

    with pytest.raises(KeyError):
        make_eff_hists.decode_arguments(
            ["--list", "nonsense"]
        )

    with pytest.raises(SystemExit):
        make_eff_hists.decode_arguments(
            ["--list", "configs"]
        )
    captured = capsys.readouterr().out
    assert captured.startswith(
        "Sample        | Magnet | Particle\n"
        "--------------|--------|----------\n"
    )


def test_make_eff_hists_with_custom_binning(config, test_path, tmp_path):
    # Save the original binnings so we can restore them later
    orig_binnings = copy.deepcopy(binning.BINNINGS)

    config.update(
        {
            "bin_vars": ["P"],
            "binning_file": str(test_path / "test_data/custom_binning.json"),
            "local_dataframe": str(test_path / "test_data/cal_test_data.csv"),
            "magnet": "up",
            "output_dir": str(tmp_path / "test_output"),
            "particle": "Pi",
            "pid_cuts": ["DLLK < 4", "DLLK<3"],
            "sample": "Turbo18",
        }
    )

    make_eff_hists.make_eff_hists(config)
    eff_histo = pd.read_pickle(
        tmp_path / "test_output/effhists-Turbo18-up-Pi-DLLK<4-P.pkl"
    )

    # These asserts might come in handy when some detail in the boost_histogram
    # implementation changes, thus failing the reference histogram comparison.
    assert eff_histo.sum(flow=False).value == pytest.approx(1.8099485065453211)
    assert eff_histo.sum(flow=True).value == pytest.approx(41.404466293590865)
    assert eff_histo[1].value == 0.9625150986552666

    # Restore original binnings for other tests
    binning.BINNINGS = orig_binnings



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

"""Module defining the default binnings and importing user-defined binnings.

TODO: The default binnings should be probably moved to a JSON file - this would
unify the default and user-defined binning approach and give users a full
example.
"""

import json
from typing import Dict, List, Union

import numpy as np
from logzero import logger as log

VALID_PARTICLES = ["Pi", "K", "P", "Mu", "e"]


def p_binning(particle: str, low: float = 3000, high: float = 100000) -> List[float]:
    """Return a binning for the momentum.

    Args:
        particle: Particle type ["Pi", "K", ...]
        low: Optional. Lowest momentum [MeV]. Defaults to 3000.
        high: Optional. Highest momentum [MeV]. Defaults to 100000.
    """
    if particle not in VALID_PARTICLES:
        log.error(f"'{particle}' is not a valid particle for P binning")
        raise KeyError

    bins: List[float] = []
    if particle in {"Pi", "K", "P", "e"}:
        # 9300 - RICH1 kaon threshold, 15600 - RICH2 kaon threshold
        bins.extend((low, 9300, 15600))
        # Uniform bin boundaries
        uniform_bins = np.linspace(19000, high, 16).tolist()
        bins.extend(uniform_bins)
    elif particle == "Mu":
        bins = [
            low,
            6000,
            8000,
            10000,
            12000,
            14500,
            17500,
            21500,
            27000,
            32000,
            40000,
            60000,
            70000,
            high,
        ]
    return bins


def eta_binning(particle: str, low: float = 1.5, high: float = 5.0) -> List[float]:
    """Return a binning for eta.

    Args:
        particle: Particle type ["Pi", "K", ...] (not used)
        low: Optional. Lowest eta.
        high: Optional. Highest eta.
    """
    return list(np.linspace(low, high, 5))


def ntracks_binning(particle: str, low: float = 0, high: float = 500) -> List[float]:
    """Return a binning for nTracks.

    Args:
        particle: Particle type ["Pi", "K", ...] (not used)
        low: Optional. Lowest nTracks.
        high: Optional. Highest nTracks.
    """
    return [low, 50, 200, 300, high]


def nspdhits_binning(particle: str, low: float = 0, high: float = 1000) -> List[float]:
    """Return a binning for nTracks.

    Args:
        particle: Particle type ["Pi", "K", ...] (not used)
        low: Optional. Lowest nTracks.
        high: Optional. Highest nTracks.
    """
    return [low, 200, 400, 600, 800, high]


def trchi2_binning(particle: str, low: float = 0.0, high: float = 3.0) -> List[float]:
    """Return a binning for track chi2.

    Args:
        particle: Particle type ["Pi", "K", ...] (not used)
        low: Optional. Lowest track chi2.
        high: Optional. Highest track chi2.
    """
    return list(np.linspace(low, high, 4))


# Dict of binnings for each track type and variable
# sourcery skip: merge-dict-assign
BINNINGS = {}

BINNINGS["Pi"] = {
    "P": {"bin_edges": p_binning("Pi")},
    "Brunel_P": {"bin_edges": p_binning("Pi")},
    "ETA": {"bin_edges": eta_binning("Pi")},
    "Brunel_ETA": {"bin_edges": eta_binning("Pi")},
    "nTracks": {"bin_edges": ntracks_binning("Pi")},
    "nTracks_Brunel": {"bin_edges": ntracks_binning("Pi")},
    "nSPDhits": {"bin_edges": nspdhits_binning("Pi")},
    "nSPDhits_Brunel": {"bin_edges": nspdhits_binning("Pi")},
    "TRCHI2NDOF": {"bin_edges": trchi2_binning("Pi")},
}

BINNINGS["K"] = {
    "P": {"bin_edges": p_binning("K")},
    "Brunel_P": {"bin_edges": p_binning("K")},
    "ETA": {"bin_edges": eta_binning("K")},
    "Brunel_ETA": {"bin_edges": eta_binning("K")},
    "nTracks": {"bin_edges": ntracks_binning("K")},
    "nTracks_Brunel": {"bin_edges": ntracks_binning("K")},
    "nSPDhits": {"bin_edges": nspdhits_binning("K")},
    "nSPDhits_Brunel": {"bin_edges": nspdhits_binning("K")},
    "TRCHI2NDOF": {"bin_edges": trchi2_binning("K")},
}

BINNINGS["Mu"] = {
    "P": {"bin_edges": p_binning("Mu")},
    "Brunel_P": {"bin_edges": p_binning("Mu")},
    "ETA": {"bin_edges": eta_binning("Mu")},
    "Brunel_ETA": {"bin_edges": eta_binning("Mu")},
    "nTracks": {"bin_edges": ntracks_binning("Mu")},
    "nTracks_Brunel": {"bin_edges": ntracks_binning("Mu")},
    "nSPDhits": {"bin_edges": nspdhits_binning("Mu")},
    "nSPDhits_Brunel": {"bin_edges": nspdhits_binning("Mu")},
    "TRCHI2NDOF": {"bin_edges": trchi2_binning("Mu")},
}

BINNINGS["P"] = {
    "P": {"bin_edges": p_binning("P")},
    "Brunel_P": {"bin_edges": p_binning("P")},
    "ETA": {"bin_edges": eta_binning("P")},
    "Brunel_ETA": {"bin_edges": eta_binning("P")},
    "nTracks": {"bin_edges": ntracks_binning("P")},
    "nTracks_Brunel": {"bin_edges": ntracks_binning("P")},
    "nSPDhits": {"bin_edges": nspdhits_binning("P")},
    "nSPDhits_Brunel": {"bin_edges": nspdhits_binning("P")},
    "TRCHI2NDOF": {"bin_edges": trchi2_binning("P")},
}

BINNINGS["e"] = {
    "P": {"bin_edges": p_binning("e")},
    "Brunel_P": {"bin_edges": p_binning("e")},
    "ETA": {"bin_edges": eta_binning("e")},
    "Brunel_ETA": {"bin_edges": eta_binning("e")},
    "nTracks": {"bin_edges": ntracks_binning("e")},
    "nTracks_Brunel": {"bin_edges": ntracks_binning("e")},
    "nSPDhits": {"bin_edges": nspdhits_binning("e")},
    "nSPDhits_Brunel": {"bin_edges": nspdhits_binning("e")},
    "TRCHI2NDOF": {"bin_edges": trchi2_binning("e")},
}


def set_binning(particle: str, variable: str, bin_edges: List[float]) -> None:
    """Set a new binning for a variable of a particle.

    Either a binning for a new particle/variable is added or the existing
    binning is rewritten.

    Args:
        particle: Particle name.
        variable: Variable name, e.g., "P" or "Brunel_ETA"
        bin_edges: A list of all bin edges.
    """
    if not isinstance(bin_edges, list):
        log.error("bin_edges parameter is not a list.")
        raise TypeError

    if particle not in BINNINGS:
        BINNINGS[particle] = {}

    BINNINGS[particle][variable] = {"bin_edges": bin_edges}


def get_binning(
    particle: str, variable: str, verbose: bool = False, quiet: bool = False
) -> List[float]:
    """Return a suitable binning for a particle and variable.

    Args:
        particle: Particle name.
        variable: Variable name, e.g., "P" or "Brunel_ETA"
        verbose: Optional. Print message when alternative binning is used.
            Defaults to False.
        quiet: Optional. Suppress all logging messages. Defaults to False.
    """
    if particle in BINNINGS and variable in BINNINGS[particle]:
        return BINNINGS[particle][variable]["bin_edges"]

    # Remove particle suffix, e.g., 'DsPhi' in 'K_DsPhi'
    pure_particle = particle.split("_", 1)[0]
    if pure_particle not in BINNINGS or variable not in BINNINGS[pure_particle]:
        if not quiet:
            log.error(f"No '{variable}' binning defined for particle {particle}")
        raise KeyError
    else:
        if not quiet and verbose:
            log.info(
                (
                    f"No '{variable}' binning defined for particle "
                    f"'{particle}'. Falling back to particle "
                    f"'{pure_particle}' binning."
                )
            )
        return BINNINGS[pure_particle][variable]["bin_edges"]


def load_binnings(path: str) -> Dict[str, Dict]:
    """Load binnings from a JSON file.

    Args:
        path: Path to the binning JSON file.

    Returns:
        A dictionary with the new binnings.
    """
    new_binnings = {}
    log.info(f"Loading binnings from {path}")
    with open(path) as f:
        new_binnings = json.load(f)
    for particle, variables in new_binnings.items():
        for variable, binning in variables.items():
            set_binning(particle, variable, binning)

    return new_binnings


def check_and_load_binnings(
    particle: str, bin_vars: List[str], binning_file: Union[str, None]
) -> None:
    """Load custom binnings and check if all necessary binnings exits.

    Args:
        particle: Particle type (K, pi, etc.).
        bin_vars: Binning variables.
        binning_file: Optional. Path to the binning JSON file.
    """
    custom_binnings = load_binnings(binning_file) if binning_file else {}
    # Check that all binnings exist
    for bin_var in bin_vars:
        bin_edges = get_binning(particle, bin_var, verbose=True)
        log.debug(f"{bin_var} binning: {bin_edges}")
        # Check if a custom binning exists and label it as used
        if particle in custom_binnings and bin_var in custom_binnings[particle]:
            custom_binnings[particle][bin_var] = "used"

    unused_custom_binnings: List[Dict[str, str]] = []
    for particle in custom_binnings:
        unused_custom_binnings.extend(
            {"particle": particle, "bin_var": bin_var}
            for bin_var in custom_binnings[particle]
            if custom_binnings[particle][bin_var] != "used"
        )

    if unused_custom_binnings:
        log.warning(
            (
                "The following custom binnings are not "
                f"being used: {unused_custom_binnings}"
            )
        )



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

"""Module that calculates the LHCb PID efficiency of a reference sample.

This module uses the histograms created by make_eff_hists to assign
efficiency to events in a reference sample supplied by the user. Adding of
efficiency to the user-supplied file requires PyROOT and is optional.

The module works in two steps:

1. Calculate the efficiency and save it as a TTree in a separate file.
2. Optionally copy the efficiency TTree to the reference file and make it a friend of
   the user's TTree.

The second step is an efficient way of "adding" the efficiency branches to
the user's TTree. The step must be requested by specifying --merge on the
command line.

Examples:
    Evaluate efficiency of a single PID cut and add it to the reference file::

        $ python -m src.pidcalib2.ref_calib --sample=Turbo18 --magnet=up \
            --ref-file=data/user_ntuple.root --output-dir=pidcalib_output \
            --bin-vars='{"P": "mom", "ETA": "Eta", "nSPDHits": "nSPDhits"}' \
            --ref-pars='{"Bach": ["K", "DLLK > 4"]}'


    Evaluate efficiency of a single PID cut and save it to
    user_ntuple_PID_eff.root without adding it to user_ntuple.root::

        $ python -m src.pidcalib2.ref_calib --sample=Turbo18 --magnet=up \
            --ref-file=data/user_ntuple.root --output-dir=pidcalib_output \
            --bin-vars='{"P": "mom", "ETA": "Eta", "nSPDHits": "nSPDhits"}' \
            --ref-pars='{"Bach": ["K", "DLLK > 4"]}'

    Evaluate efficiency of multiple PID cuts and add them to the reference
    file::

        $ python -m src.pidcalib2.ref_calib --sample=Turbo18 --magnet=up \
            --ref-file=data/user_ntuple.root --output-dir=pidcalib_output \
            --bin-vars='{"P": "P", "ETA": "ETA", "nSPDHits": "nSPDHits"}' \
            --ref-pars='{"Bach": ["K", "DLLK > 4"], "SPi": ["Pi", "DLLK < 0"]}'
"""

import argparse
import ast
import logging
import pathlib
import sys
import time
from typing import Dict, List

import logzero
from logzero import logger as log

from pidcalib2 import argparse_tools, pid_data, utils

try:
    from pidcalib2 import merge_trees
except ModuleNotFoundError:
    log.warning("PyROOT not found; --merge won't work")

try:
    from pidcalib2.version import version as VERSION  # type: ignore
except ImportError:
    VERSION = "N/A"


def decode_arguments(args: List[str]) -> argparse.Namespace:
    """Decode CLI arguments."""
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        formatter_class=argparse_tools.RawDefaultsFormatter,
        description=(
            "This module uses the histograms created by make_eff_hists to "
            "assign efficiency to events in a reference sample supplied by "
            "the user. Adding efficiency to the user-supplied file requires "
            "PyROOT and is optional."
            "\n\n"
            "The module works in two steps:\n"
            "  1. Calculate the efficiency and save it as a TTree in a "
            "separate file.\n"
            "  2. Optionally copy the efficiency TTree to the reference file "
            "and make it a friend of the user's TTree. The user must request "
            "the step by specifying --merge on the command line.\n\n"
            "Be aware that --merge will modify your file. Use with caution."
        ),
    )
    parser.add_argument(
        "-s",
        "--sample",
        help="calibration sample (Turbo18, Electron16, ...)",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-m",
        "--magnet",
        help="magnet polarity",
        required=True,
        choices=["up", "down"],
    )
    parser.add_argument(
        "-b",
        "--bin-vars",
        help=(
            "dictionary of binning variables (keys) and their associated names in "
            "the reference sample (values)"
        ),
        default="{'P' : 'P', 'ETA' : 'ETA', 'nTracks' : 'nTracks'}",
        dest="bin_vars",
    )
    parser.add_argument(
        "-i",
        "--histo-dir",
        default="pidcalib_output",
        help="directory where efficiency histograms are located",
    )
    parser.add_argument(
        "-o",
        "--output-file",
        help="ROOT filename into which to save the PID efficiency tree",
        required=True,
    )
    parser.add_argument(
        "-f",
        "--ref-file",
        help="reference sample file",
        required=True,
    )
    parser.add_argument(
        "-t",
        "--ref-tree",
        help="reference sample tree name",
        default="DecayTree",
    )
    parser.add_argument(
        "-p",
        "--ref-pars",
        help=(
            "JSON dictionary of particles from reference sample to apply "
            "cuts to, where the keys represent the particle branch name prefix, "
            "and the values passed are a list containing particle type and "
            "PID cut e.g. \"{'D0_K' : ['K','DLLK>4'], 'D0_Pi' : ['Pi','DLLK<4']}\""
        ),
        required=True,
    )
    parser.add_argument(
        "-r",
        "--merge",
        help="merge the PID efficiency tree with the reference file",
        action="store_true",
    )
    parser.add_argument(
        "-c",
        "--compatibility",
        action="store_true",
        help="(debug) treat empty efficiency histogram bins as PIDCalib1 did",
    )
    parser.add_argument(
        "-l",
        "--list",
        action=argparse_tools.ListValidAction,
        help="list all [configs, aliases]",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="(debug) increase verbosity",
    )
    parser.add_argument("-V", "--version", action="version", version=VERSION)
    return parser.parse_args(args)


def ref_calib(config: Dict) -> float:
    """Assign efficiency to tracks and events in a user-supplied dataset.

    Each track that falls within the phasespace covered by the efficiency
    histogram is assigned efficiency of the bin it falls into. The overall
    event efficiency is a product of all the track efficiencies. The
    resulting efficiency columns/branches are saved to a TTree in a file
    <reference_filename>_PID_eff.root. The TTree is then copied to the
    reference (user) file and made a friend TTree of the original TTree
    specified by the user. This allows the user to treat their original TTree
    as if it itself had the efficiency branches.

    Args:
        config: A configuration dictionary. See decode_arguments(args) for
            details.

    Returns:
        Average efficiency of all the events.
    """
    if config["verbose"]:
        logzero.loglevel(logging.DEBUG)
    else:
        logzero.loglevel(logging.INFO)

    config["version"] = VERSION
    log.info("Running PIDCalib2 ref_calib with the following config:")
    utils.log_config(config)

    try:
        bin_vars = ast.literal_eval(config["bin_vars"])
        if not isinstance(bin_vars, dict):
            raise SyntaxError
    except SyntaxError:
        log.error("The --bin-vars string is not a valid Python dict")
        raise

    try:
        ref_pars = ast.literal_eval(config["ref_pars"])
        if not isinstance(ref_pars, dict):
            raise SyntaxError
    except SyntaxError:
        log.error("The --ref-pars string is not valid Python dict")
        raise

    eff_histos = pid_data.get_calib_hists(
        config["histo_dir"], config["sample"], config["magnet"], ref_pars, bin_vars
    )

    log.info(f"Loading reference sample '{config['ref_file']}' ...")
    ref_branches = pid_data.get_reference_branch_names(ref_pars, bin_vars)
    df_ref = pid_data.root_to_dataframe(
        config["ref_file"], [config["ref_tree"]], ref_branches
    )
    if df_ref is None:
        log.error("Failed to load reference sample")
        raise IOError
    log.debug(
        f"Reference sample '{config['ref_file']}' with {len(df_ref.index)} events loaded"  # noqa
    )

    start = time.perf_counter()
    df_ref = utils.add_bin_indices(df_ref, list(ref_pars), bin_vars, eff_histos)
    df_ref = utils.add_efficiencies(df_ref, list(ref_pars), eff_histos)
    end = time.perf_counter()
    log.debug(f"Efficiency calculation took {end-start:.2f}s")

    # Calculate average of the per-event effs
    # Use only data with valid eff values (those events falling inside the
    # calibration hist)
    avg_eff = df_ref["PIDCalibEff"].dropna().mean()
    log.info(f"Average per-event PID efficiency: {avg_eff:.2%}")

    output_path = pathlib.Path(config["output_file"])
    output_path.parent.mkdir(parents=True, exist_ok=True)

    pid_data.save_dataframe_as_root(
        df_ref[[key for key in df_ref.keys() if key not in ref_branches]],
        "PIDCalibTree",
        str(output_path),
    )

    if config["merge"]:
        merge_trees.copy_tree_and_set_as_friend(
            str(output_path), "PIDCalibTree", config["ref_file"], config["ref_tree"]
        )

    return avg_eff


def main():  # sourcery skip: docstrings-for-functions, require-return-annotation
    config = vars(decode_arguments(sys.argv[1:]))
    ref_calib(config)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import os
from pathlib import Path

import pytest

from pidcalib2 import ref_calib


@pytest.fixture
def test_path():
    return Path(os.path.dirname(os.path.abspath(__file__)))


def test_ref_calib(test_path, tmp_path):
    config = {
        "sample": "Turbo18",
        "magnet": "up",
        "bin_vars": '{"P": "P", "ETA": "ETA", "nTracks": "nTracks"}',
        "histo_dir": str(test_path / "test_data"),
        "output_file": str(tmp_path / "PIDCalibResults.root"),
        "merge": False,
        "ref_file": str(test_path / "test_data/ref_test_data.root"),
        "ref_tree": "DecayTree",
        "ref_pars": '{"Bach": ["K", "DLLK > 4"]}',
        "verbose": False,
    }

    assert ref_calib.ref_calib(config) == pytest.approx(0.8798221261720731)
    assert (tmp_path / "PIDCalibResults.root").exists()

    with pytest.raises(SyntaxError):
        config = {
            "sample": "Turbo18",
            "magnet": "up",
            "bin_vars": '{"P", "ETA", "nTracks"}',
            "output_file": str(tmp_path / "test_data/test_data_PIDCalibResults.root"),
            "merge": False,
            "ref_file": str(test_path / "test_data/ref_test_data.root"),
            "ref_tree": "DecayTree",
            "ref_pars": '{"Bach": ["K", "DLLK > 4"]}',
            "verbose": False,
        }
        ref_calib.ref_calib(config)

    with pytest.raises(SyntaxError):
        config = {
            "sample": "Turbo18",
            "magnet": "up",
            "bin_vars": '{"P": "P", "ETA": "ETA", "nTracks": "nTracks"}',
            "output_file": str(tmp_path / "test_data/test_data_PIDCalibResults.root"),
            "merge": False,
            "ref_file": str(test_path / "test_data/ref_test_data.root"),
            "ref_tree": "DecayTree",
            "ref_pars": '{"Bach"}',
            "verbose": False,
        }
        ref_calib.ref_calib(config)


def test_decode_arguments():
    assert (
        ref_calib.decode_arguments(
            [
                "--sample=Turbo18",
                "-m=up",
                "--ref-file=ref_file.root",
                "--ref-pars={'Ref': ['pi', 'DLLK<5']}",
                "--output-file=output.root",
            ]
        ).magnet
        == "up"
    )

    with pytest.raises(SystemExit):
        ref_calib.decode_arguments(
            [
                "--sample=Turbo18",
                "-m=up",
                "--ref-file=ref_file.root",
                "--ref-pars={'Ref': ['pi', 'DLLK<5']}",
            ]
        ).magnet

    with pytest.raises(SystemExit):
        ref_calib.decode_arguments(
            [
                "--sample=Turbo18",
                "-m=up",
                "--ref-file=ref_file.root",
                "--ref-pars={'Ref': ['pi', 'DLLK<5']}",
                "--particle=K",
            ]
        )

    with pytest.raises(SystemExit):
        ref_calib.decode_arguments(
            [
                "--sample=Turbo18",
                "-m=MagUp",
                "--ref-file=ref_file.root",
                "--ref-pars={'Ref': ['pi', 'DLLK<5']}",
            ]
        )

    with pytest.raises(SystemExit):
        ref_calib.decode_arguments(
            [
                "--sample=Turbo18",
                "-m=up",
                "--ref-file=ref_file.root",
            ]
        )



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import os
import shutil
from pathlib import Path

import pytest
try:
    import ROOT
    from pidcalib2 import merge_trees
except ImportError:
    pytest.skip("PyROOT not available", allow_module_level=True)


@pytest.fixture
def test_path():
    return Path(os.path.dirname(os.path.abspath(__file__)))


@pytest.mark.pyroot
def test_copy_tree(test_path, tmp_path):
    shutil.copy(
        test_path / "test_data/ref_test_data.root",
        tmp_path / "ref_test_data_copy.root",
    )

    merge_trees.copy_tree_and_set_as_friend(
        str(test_path / "test_data/ref_PID_eff.root"),
        "PID_eff_tree",
        str(tmp_path / "ref_test_data_copy.root"),
        "DecayTree",
    )

    new_file = ROOT.TFile(str(tmp_path / "ref_test_data_copy.root"))
    new_tree = new_file.Get("DecayTree")

    assert new_tree.GetEntry(0) == 84
    assert new_tree.PID_eff == pytest.approx(0.9933816230604905)


@pytest.mark.pyroot
def test_copy_tree_with_subdir(test_path, tmp_path):
    shutil.copy(
        test_path / "test_data/ref_test_data_subdir.root",
        tmp_path / "ref_test_data_subdir_copy.root",
    )

    merge_trees.copy_tree_and_set_as_friend(
        str(test_path / "test_data/ref_PID_eff.root"),
        "PID_eff_tree",
        str(tmp_path / "ref_test_data_subdir_copy.root"),
        "subdir/DecayTree",
    )

    new_file = ROOT.TFile(str(tmp_path / "ref_test_data_subdir_copy.root"))
    new_tree = new_file.Get("subdir/DecayTree")

    # Check that the tree is not in the ROOT directory
    with pytest.raises(ReferenceError):
        new_file.Get("DecayTree").GetName()

    assert new_tree.GetEntry(0) == 84
    assert new_tree.PID_eff == pytest.approx(0.9933816230604905)


@pytest.mark.pyroot
def test_copy_tree_bad_filenames(test_path, tmp_path):
    shutil.copy(
        test_path / "test_data/ref_test_data.root",
        tmp_path / "ref_test_data_copy.root",
    )
    # It seems some versions of pyroot raise OSError when a file that should be
    # opened doesn't exist, while others don't and we catch the issue later,
    # raising a SystemExit
    with pytest.raises((OSError, SystemExit)):
        merge_trees.copy_tree_and_set_as_friend(
            str(test_path / "test_data/x.root"),
            "PID_eff_tree",
            str(tmp_path / "ref_test_data_copy.root"),
            "DecayTree",
        )

    with pytest.raises(SystemExit):
        merge_trees.copy_tree_and_set_as_friend(
            str(test_path / "test_data/ref_PID_eff.root"),
            "PID_eff_tree",
            str(test_path / "test_data/x.root"),
            "DecayTree",
        )


@pytest.mark.pyroot
def test_copy_tree_bad_treenames(test_path, tmp_path):
    shutil.copy(
        test_path / "test_data/ref_test_data.root",
        tmp_path / "ref_test_data_copy.root",
    )

    with pytest.raises(SystemExit):
        merge_trees.copy_tree_and_set_as_friend(
            str(test_path / "test_data/ref_PID_eff.root"),
            "x",
            str(tmp_path / "ref_test_data_copy.root"),
            "DecayTree",
        )

    with pytest.raises(SystemExit):
        merge_trees.copy_tree_and_set_as_friend(
            str(test_path / "test_data/ref_PID_eff.root"),
            "PID_eff_tree",
            str(tmp_path / "ref_test_data_copy.root"),
            "x",
        )



"""A set of software tools for estimating LHCb PID efficiencies.

The package includes several user-callable modules:
- make_eff_hists creates histograms that can be used to estimate the PID
  efficiency of a user's sample
- ref_calib calculates the LHCb PID efficiency of a user reference sample
- merge_trees merges two ROOT files with compatible TTrees
- plot_calib_distributions allows you to plot distributions of variables in the
  calibration datasets
- pklhisto2root converts Pickled boost-histograms to ROOT histograms

More information can be found in README.md.
"""



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

"""Module defining tree names in the PID calibration files."""

TUPLE_NAMES = {
    "Pi": ["DSt_PiMTuple", "DSt_PiPTuple"],
    "K": ["DSt_KPTuple", "DSt_KMTuple"],
    "Mu": ["Jpsi_MuPTuple", "Jpsi_MuMTuple"],
    "K_DsPhi": ["DsPhi_KPTuple", "DsPhi_KMTuple"],
    "K_DSt3Pi": ["DSt3Pi_KPTuple", "DSt3Pi_KMTuple"],
    "Pi_DSt3Pi": ["DSt3Pi_PiMTuple", "DSt3Pi_PiPTuple"],
    "K_Phi": ["Phi_KMTuple", "Phi_KPTuple"],
    "K_Omega": ["OmegaL_KPTuple", "OmegaL_KMTuple"],
    "K_DD": ["OmegaDDD_KPTuple", "OmegaDDD_KMTuple"],
    "Pi_KS": ["KSLL_PiPTuple", "KSLL_PiMTuple"],
    "Pi_DD": ["KSDD_PiPTuple", "KSDD_PiMTuple"],
    "P": [
        "Lam0LL_PTuple",
        "Lam0LL_PbarTuple",
        "Lam0LL_HPT_PTuple",
        "Lam0LL_HPT_PbarTuple",
        "Lam0LL_VHPT_PTuple",
        "Lam0LL_VHPT_PbarTuple",
    ],
    "P_DD": [
        "Lam0DD_PTuple",
        "Lam0DD_PbarTuple",
        "Lam0DD_HPT_PTuple",
        "Lam0DD_HPT_PbarTuple",
        "Lam0DD_VHPT_PTuple",
        "Lam0DD_VHPT_PbarTuple",
    ],
    "P_Sigmac0": ["Sigmac0_PbarTuple", "Sigmac0_PTuple"],
    "P_Sigmacpp": ["Sigmacpp_PbarTuple", "Sigmacpp_PTuple"],
    "P_B_Jpsi": ["B_Jpsi_PTuple", "B_Jpsi_PbarTuple"],
    "Mu_B_Jpsi": ["B_Jpsi_MuPTuple", "B_Jpsi_MuMTuple"],
    "e": ["Jpsi_EPTuple", "Jpsi_EMTuple"],
    "P_LcfB": ["LbLcMu_PTuple", "LbLcMu_PbarTuple"],
    "P_IncLc": ["Lc_PTuple", "Lc_PbarTuple"],
    "Mu_nopt": ["Jpsinopt_MuPTuple", "Jpsinopt_MuMTuple"],
}

SIMPLE_SAMPLES = [
    "13b",
    "15",
    "17",
    "20",
    "20_MCTuneV2",
    "20_MCTunev3",
    "20r1",
    "20r1_MCTuneV2",
    "21",
    "21_MCTuneV4",
    "21r1",
    "21r1_MCTuneV4",
    "22",
    "23",
    "23Val",
    "23_MCTuneV1",
    "26",
    "5TeV",
    "Electron15",
    "Electron16",
    "Electron17",
    "Electron18",
]



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

"""Module that contains contains overrides for argparse.

It allows us to print the valid configs and aliases when the user requests it,
as well as print the default values for parameters together with a
RawDecription (allowing more control over formatting).
"""

import argparse
import contextlib
from logzero import logger as log

from pidcalib2 import markdown_table, pid_data


class ListValidAction(argparse.Action):
    """Class that overrides required parameters and prints valid configs."""

    def __call__(self, parser, namespace, values, option_string=None):
        # sourcery skip: docstrings-for-functions

        if values == "configs" or values.endswith(".json"):
            header = ["Sample", "Magnet", "Particle"]
            table = markdown_table.MarkdownTable(header)

            # Print configs from the default samples.json if no file specified
            if values == "configs":
                values = None

            for entry in pid_data.get_calibration_samples(values).keys():
                # Skip group entries like "Turbo18-MagUp"
                with contextlib.suppress(ValueError):
                    sample, magnet, particle = entry.split("-")
                    magnet = magnet[3:].lower()
                    table.add_row([sample, magnet, particle])
            table.print()

        elif values == "aliases":
            table_pid = markdown_table.MarkdownTable(["Alias", "Variable"])
            for alias, var in pid_data.ALIASES.items():
                table_pid.add_row([alias, var])
            table_pid.print()
        else:
            log.error(f"'{values}' is not a known keyword for list-valid")
            raise KeyError

        parser.exit()


class RawDefaultsFormatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter
):
    """Combines the ArgumentDefaults and the RawDescription formatters."""
    pass



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

"""Module to make plots of LHCb PID calibration sample variables.

Examples:
    Create plots of the variables DLLK and P using 1 calibration file::

        $ python -m src.pidcalib2.plot_calib_distributions --sample=Turbo18 \
            --magnet=up --particle=Pi --bin-var=DLLK --bin-var=P \
            --output-dir=pidcalib_output --max-files=1

    Create PDF plots of variable P with 95 uniform bins using 1 calib. file::

        $ python -m src.pidcalib2.plot_calib_distributions --sample=Turbo18 \
            --magnet=up --particle=Pi --bin-var=P  --output-dir=pidcalib_output \
            --max-files=1 --format=pdf --force-uniform --bins=95

    Create plots of variable P using custom binning and 1 calib. file::

        $ python -m src.pidcalib2.plot_calib_distributions --sample=Turbo18 \
            --magnet=up --particle=Pi --bin-var=P  --output-dir=pidcalib_output \
            --max-files=1 --format=png --binning-file=my_binning.json
"""

import argparse
import logging
import pathlib
import pickle
import sys
from typing import Dict, List

import boost_histogram as bh
import logzero
import matplotlib
import mplhep
import numpy as np
from logzero import logger as log
from matplotlib import pyplot as plt
from tqdm import tqdm

from pidcalib2 import argparse_tools, binning, pid_data, utils

try:
    from pidcalib2.version import version as VERSION  # type: ignore
except ImportError:
    VERSION = "N/A"

# Avoid matplotlib trying to use QT based backends
matplotlib.use("agg")


def decode_arguments(args: List[str]) -> argparse.Namespace:
    """Decode CLI arguments."""
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        formatter_class=argparse_tools.RawDefaultsFormatter,
        description=(
            "This tool allows you to plot distributions of variables in the "
            "calibration datasets. You can supply the same cuts and custom "
            "binnings that you would use for make_eff_hists. If you wish to plot "
            "a variable for which no binning exists, a uniform binning with 50 "
            "bins will be used. You can change the number of bins using --bins "
            "and force a uniform binning even if another binning is defined via "
            "--force-uniform."
            "\n\n"
            "A plot for every requested variable will be "
            "created in the --output-dir directory. The format of the plots can "
            "be controlled by --format. Furthermore, "
            "plot_calib_distributions.pkl will be saved in the same directory, "
            "containing all the histograms, should the user want to make the "
            "plots manually."
        ),
    )
    parser.add_argument(
        "-s",
        "--sample",
        help="calibration sample (Turbo18, Electron16, ...)",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-m",
        "--magnet",
        help="magnet polarity",
        required=True,
        choices=["up", "down"],
    )
    parser.add_argument(
        "-p",
        "--particle",
        help="particle type; see pidcalib2.plot_calib_distributions --list configs",
        required=True,
    )
    parser.add_argument(
        "-b",
        "--bin-var",
        dest="bin_vars",
        help="variable to plot (-b can be specified multiple times)",
        action="append",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        help="directory where to save output files",
        default="pidcalib_output",
    )
    parser.add_argument(
        "-r",
        "--format",
        help="format of the plots",
        choices=["png", "pdf"],
        default="png",
    )
    parser.add_argument(
        "-g",
        "--binning-file",
        help="file where new/alternative binnings are defines",
    )
    parser.add_argument(
        "-a",
        "--samples-file",
        help="(debug) read calibration sample lists from a custom file",
    )
    parser.add_argument(
        "-i",
        "--bins",
        help="number of uniform bins to use when no binning exists",
        default=50,
        type=int,
    )
    parser.add_argument(
        "-u",
        "--force-uniform",
        help="force the usage of uniform binning even if other binning exists",
        action="store_true",
    )
    parser.add_argument(
        "-c",
        "--cut",
        dest="cuts",
        help=(
            "arbitrary cut string, e.g., 'Dst_IPCHI2 < 10.0' (-c can be used multiple "
            "times for multiple cuts)."
        ),
        action="append",
    )
    parser.add_argument(
        "-l",
        "--list",
        action=argparse_tools.ListValidAction,
        help="list all [configs, aliases]",
    )
    parser.add_argument(
        "-f",
        "--file-list",
        help="(debug) read calibration file paths from a text file",
    )
    parser.add_argument(
        "-n",
        "--max-files",
        type=int,
        help="(debug) a max number of files to read",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="(debug) increase verbosity",
    )
    parser.add_argument("-V", "--version", action="version", version=VERSION)
    return parser.parse_args(args)


def plot_calib_distributions(config: Dict) -> None:
    """Create, plot, and save histograms of a calibration sample.

    The specified calibration sample is processed with cuts taken into account.
    The requested variables are histogrammed, possibly with custom binning
    (specified in the same way as for make_eff_hists).

    Args:
        config: A configuration dictionary. See decode_arguments(args) for
            details.
    """
    if config["verbose"]:
        logzero.loglevel(logging.DEBUG)
    else:
        logzero.loglevel(logging.INFO)

    config["version"] = VERSION
    log.info("Running PIDCalib2 plot_calib_distributions with the following config:")
    utils.log_config(config)

    output_dir = pathlib.Path(config["output_dir"])
    output_dir.mkdir(parents=True, exist_ok=True)

    calib_sample = {}
    if config["file_list"]:
        with open(config["file_list"]) as f_list:
            calib_sample["files"] = f_list.read().splitlines()
    else:
        calib_sample = pid_data.get_calibration_sample(
            config["sample"],
            config["magnet"],
            config["particle"],
            config["samples_file"],
            config["max_files"],
        )
    tree_paths = pid_data.get_tree_paths(config["particle"], config["sample"])
    log.debug(f"Trees to be read: {tree_paths}")

    # If there are hard-coded cuts, the variables must be included in the
    # branches to read.
    cuts = config["cuts"]
    if "cuts" in calib_sample:
        if cuts is None:
            cuts = []
        cuts += calib_sample["cuts"]

    branch_names = pid_data.get_relevant_branch_names([], config["bin_vars"], cuts)
    log.debug(f"Branches to be read: {branch_names}")
    log.info(
        f"{len(calib_sample['files'])} calibration files from EOS will be processed"
    )
    for path in calib_sample["files"]:
        log.debug(f"  {path}")

    if config["binning_file"]:
        binning.load_binnings(config["binning_file"])

    if config["force_uniform"]:
        log.info("Deleting binnings to force uniform binning")
        binning.BINNINGS = {}

    all_hists = create_plot_histograms(config, calib_sample, tree_paths, branch_names)
    total_hists = utils.add_hists(list(all_hists.values()))

    save_histograms(total_hists, output_dir)
    save_plots(total_hists, output_dir, config["format"])


def create_plot_histograms(
    config: Dict,
    calib_sample: Dict,
    tree_paths: List[str],
    branch_names: Dict[str, str],
) -> Dict[str, Dict[str, bh.Histogram]]:
    """Create histograms for plotting.

    A separate set of histograms is created for each file in the calib_sample.
    The set comprises histograms for each variable in branch_names.

    Args:
        config: A configuration dictionary. See decode_arguments(args) for
            details.
        calib_sample: Calibration sample - files, cuts, etc.
        tree_paths: List of internal ROOT paths to relevant trees in the files.
        branch_names: Dict of branch names {user-level name: branch name}.

    Returns:
        A dictionary {file: {var: histogram}}.
    """
    bin_vars_without_binnings = []
    binning_range_cuts = []
    for bin_var in config["bin_vars"]:
        try:
            bin_edges = binning.get_binning(
                config["particle"], bin_var, verbose=True, quiet=True
            )
            binning_range_cuts.append(
                f"{bin_var} > {bin_edges[0]} and {bin_var} < {bin_edges[-1]}"
            )
        except KeyError:
            log.warning(
                (
                    f"No '{bin_var}' binning defined for particle "
                    f"{config['particle']}; will use {config['bins']} uniform bins"
                )
            )
            bin_vars_without_binnings.append(bin_var)

    cut_stats = {
        "binning range": {"before": 0, "after": 0},
        "hard-coded": {"before": 0, "after": 0},
        "user": {"before": 0, "after": 0},
    }

    all_hists = {}

    for path in (
        tqdm(calib_sample["files"], leave=False, desc="Processing files")
        if sys.stderr.isatty()  # Use tqdm only when running interactively
        else calib_sample["files"]
    ):
        df = pid_data.root_to_dataframe(
            path, tree_paths, list(branch_names.values()), True
        )
        if df is not None:
            # Rename colums of the dataset from branch names to simple user-level
            # names, e.g., probe_PIDK -> DLLK.
            inverse_branch_names = {val: key for key, val in branch_names.items()}
            df = df.rename(columns=inverse_branch_names)

            utils.apply_all_cuts(
                df,
                cut_stats,
                binning_range_cuts,
                calib_sample["cuts"] if "cuts" in calib_sample else [],
                config["cuts"] if "cuts" in config else [],
            )

            # If no binning
            for var in bin_vars_without_binnings:
                var_range = df[var].max() - df[var].min()
                var_min = df[var].min() - 0.1 * var_range
                var_max = df[var].max() + 0.1 * var_range
                binning.set_binning(
                    config["particle"],
                    var,
                    list(np.linspace(var_min, var_max, config["bins"] + 1)),
                )
            # Empty the list to avoid redefining the binning in each step
            bin_vars_without_binnings = []

            hists = {
                var: utils.make_hist(df, config["particle"], [var])
                for var in config["bin_vars"]
            }

            all_hists[path] = hists

    log.info(f"Processed {len(all_hists)}/{len(calib_sample['files'])} files")
    utils.print_cut_summary(cut_stats)

    return all_hists


def save_plots(
    total_hists: Dict[str, bh.Histogram], output_dir: pathlib.Path, format: str
) -> None:
    """Create and save plots of the supplied 1D histograms.

    Args:
        total_hists: A dictionary of the histograms to be plotted.
        output_dir: Directory to which to save the plots.
        format: Image format, e.g., png.
    """
    try:
        # New mplhep style setting
        mplhep.style.use("LHCb2")
    except AttributeError:
        # Fallback to old mplhep style setting
        mplhep.set_style("LHCb2")

    for var, hist in total_hists.items():
        plt.hist(
            hist.axes[0].edges[:-1],
            bins=hist.axes[0].edges,
            weights=hist.values(),
            histtype="step",
        )
        plt.ylabel("Events")
        plt.xlabel(var)
        plt.tight_layout()
        path = output_dir / pathlib.Path(f"{var}.{format}")
        log.info(f"Saving plot {path}")
        plt.savefig(path, dpi=100)
        plt.close()


def save_histograms(
    total_hists: Dict[str, bh.Histogram], output_dir: pathlib.Path
) -> None:
    """Pickle histograms and save them to a file.

    Args:
        total_hists: A dictionary of the histograms to be plotted.
        output_dir: Directory to which to save the histograms.
    """
    path = output_dir / pathlib.Path("plot_calib_distributions.pkl")
    log.info(f"Saving histograms to {path}")
    with open(path, "wb") as f:
        for hist in total_hists.values():
            pickle.dump(hist, f)


def main():  # sourcery skip: docstrings-for-functions, require-return-annotation
    config = vars(decode_arguments(sys.argv[1:]))
    plot_calib_distributions(config)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

"""Module with PID aliases - shorthand names for the actual PID variables."""

ALIASES = {
    "Brunel_DLLK": "probe_Brunel_PIDK",
    "Brunel_DLLd": "probe_Brunel_PIDd",
    "Brunel_DLLe": "probe_Brunel_PIDe",
    "Brunel_DLLmu": "probe_Brunel_PIDmu",
    "Brunel_DLLp": "probe_Brunel_PIDp",
    "Brunel_ETA": "probe_Brunel_ETA",
    "Brunel_HasRich": "probe_Brunel_hasRich",
    "Brunel_IPCHI2": "probe_Brunel_IPCHI2",
    "Brunel_InMuonAcc": "probe_Brunel_InMuonAcc",
    "Brunel_IsMuon": "probe_Brunel_isMuon",
    "Brunel_MC12TuneV2_ProbNNe": "probe_Brunel_MC12TuneV2_ProbNNe",
    "Brunel_MC12TuneV2_ProbNNghost": "probe_Brunel_MC12TuneV2_ProbNNghost",
    "Brunel_MC12TuneV2_ProbNNk": "probe_Brunel_MC12TuneV2_ProbNNk",
    "Brunel_MC12TuneV2_ProbNNmu": "probe_Brunel_MC12TuneV2_ProbNNmu",
    "Brunel_MC12TuneV2_ProbNNp": "probe_Brunel_MC12TuneV2_ProbNNp",
    "Brunel_MC12TuneV2_ProbNNpi": "probe_Brunel_MC12TuneV2_ProbNNpi",
    "Brunel_MC12TuneV3_ProbNNe": "probe_Brunel_MC12TuneV3_ProbNNe",
    "Brunel_MC12TuneV3_ProbNNghost": "probe_Brunel_MC12TuneV3_ProbNNghost",
    "Brunel_MC12TuneV3_ProbNNk": "probe_Brunel_MC12TuneV3_ProbNNk",
    "Brunel_MC12TuneV3_ProbNNmu": "probe_Brunel_MC12TuneV3_ProbNNmu",
    "Brunel_MC12TuneV3_ProbNNp": "probe_Brunel_MC12TuneV3_ProbNNp",
    "Brunel_MC12TuneV3_ProbNNpi": "probe_Brunel_MC12TuneV3_ProbNNpi",
    "Brunel_MC12TuneV4_ProbNNe": "probe_Brunel_MC12TuneV4_ProbNNe",
    "Brunel_MC12TuneV4_ProbNNghost": "probe_Brunel_MC12TuneV4_ProbNNghost",
    "Brunel_MC12TuneV4_ProbNNk": "probe_Brunel_MC12TuneV4_ProbNNk",
    "Brunel_MC12TuneV4_ProbNNmu": "probe_Brunel_MC12TuneV4_ProbNNmu",
    "Brunel_MC12TuneV4_ProbNNp": "probe_Brunel_MC12TuneV4_ProbNNp",
    "Brunel_MC12TuneV4_ProbNNpi": "probe_Brunel_MC12TuneV4_ProbNNpi",
    "Brunel_MC15TuneV1_ProbNNe": "probe_Brunel_MC15TuneV1_ProbNNe",
    "Brunel_MC15TuneV1_ProbNNghost": "probe_Brunel_MC15TuneV1_ProbNNghost",
    "Brunel_MC15TuneV1_ProbNNk": "probe_Brunel_MC15TuneV1_ProbNNk",
    "Brunel_MC15TuneV1_ProbNNmu": "probe_Brunel_MC15TuneV1_ProbNNmu",
    "Brunel_MC15TuneV1_ProbNNp": "probe_Brunel_MC15TuneV1_ProbNNp",
    "Brunel_MC15TuneV1_ProbNNpi": "probe_Brunel_MC15TuneV1_ProbNNpi",
    "Brunel_MuonUnbiased": "probe_Brunel_MuonUnbiased",
    "Brunel_P": "probe_Brunel_P",
    "Brunel_PT": "probe_Brunel_PT",
    "Brunel_TRACK_GHOSTPROB": "probe_Brunel_TRACK_GHOSTPROB",
    "Brunel_TRCHI2NDOF": "probe_Brunel_TRCHI2NDOF",
    "DLLK": "probe_PIDK",
    "DLLd": "probe_PIDd",
    "DLLe": "probe_PIDe",
    "DLLmu": "probe_PIDmu",
    "DLLp": "probe_PIDp",
    "ETA": "probe_ETA",
    "HasRich": "probe_hasRich",
    "IPCHI2": "probe_IPCHI2",
    "InMuonAcc": "probe_InMuonAcc",
    "IsMuon": "probe_isMuon",
    "MC12TuneV2_ProbNNe": "probe_MC12TuneV2_ProbNNe",
    "MC12TuneV2_ProbNNghost": "probe_MC12TuneV2_ProbNNghost",
    "MC12TuneV2_ProbNNk": "probe_MC12TuneV2_ProbNNk",
    "MC12TuneV2_ProbNNmu": "probe_MC12TuneV2_ProbNNmu",
    "MC12TuneV2_ProbNNp": "probe_MC12TuneV2_ProbNNp",
    "MC12TuneV2_ProbNNpi": "probe_MC12TuneV2_ProbNNpi",
    "MC12TuneV3_ProbNNe": "probe_MC12TuneV3_ProbNNe",
    "MC12TuneV3_ProbNNghost": "probe_MC12TuneV3_ProbNNghost",
    "MC12TuneV3_ProbNNk": "probe_MC12TuneV3_ProbNNk",
    "MC12TuneV3_ProbNNmu": "probe_MC12TuneV3_ProbNNmu",
    "MC12TuneV3_ProbNNp": "probe_MC12TuneV3_ProbNNp",
    "MC12TuneV3_ProbNNpi": "probe_MC12TuneV3_ProbNNpi",
    "MC12TuneV4_ProbNNe": "probe_MC12TuneV4_ProbNNe",
    "MC12TuneV4_ProbNNghost": "probe_MC12TuneV4_ProbNNghost",
    "MC12TuneV4_ProbNNk": "probe_MC12TuneV4_ProbNNk",
    "MC12TuneV4_ProbNNmu": "probe_MC12TuneV4_ProbNNmu",
    "MC12TuneV4_ProbNNp": "probe_MC12TuneV4_ProbNNp",
    "MC12TuneV4_ProbNNpi": "probe_MC12TuneV4_ProbNNpi",
    "MC15TuneV1_ProbNNe": "probe_MC15TuneV1_ProbNNe",
    "MC15TuneV1_ProbNNghost": "probe_MC15TuneV1_ProbNNghost",
    "MC15TuneV1_ProbNNk": "probe_MC15TuneV1_ProbNNk",
    "MC15TuneV1_ProbNNmu": "probe_MC15TuneV1_ProbNNmu",
    "MC15TuneV1_ProbNNp": "probe_MC15TuneV1_ProbNNp",
    "MC15TuneV1_ProbNNpi": "probe_MC15TuneV1_ProbNNpi",
    "MuonUnbiased": "probe_MuonUnbiased",
    "P": "probe_P",
    "PT": "probe_PT",
    "TRACK_GHOSTPROB": "probe_TRACK_GHOSTPROB",
    "TRCHI2NDOF": "probe_TRCHI2NDOF",
    "nSPDhits": "nSPDhits",
    "nSPDhits_Brunel": "nSPDhits_Brunel",
    "nTracks": "nTracks",
    "nTracks_Brunel": "nTracks_Brunel",
    "trackcharge": "probe_trackcharge",
}



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

"""This module prints an error message when calling pidcalib2 directly.

The pidcalib2 package is not meant to be called directly - one has to call one
of the work modules, e.g., pidcalib2.make_eff_hists. More info in README.md.
"""
import sys

if __name__ == "__main__":
    print("You must call one of the work modules, e.g., pidcalib2.make_eff_hists.")
    sys.exit(1)



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import pickle

import boost_histogram as bh
import numpy as np
import pytest

from pidcalib2 import binning, plot_calib_distributions


@pytest.mark.xrootd
@pytest.mark.slow
def test_plot_calib_distributions(tmp_path):
    config = {
        "particle": "K",
        "magnet": "down",
        "sample": "Turbo18",
        "max_files": 1,
        "bin_vars": ["DLLK"],
        "bins": 50,
        "force_uniform": True,
        "verbose": True,
        "output_dir": tmp_path,
        "file_list": None,
        "samples_file": None,
        "binning_file": None,
        "format": "png",
        "cuts": ["P > 10000"],
    }
    # Save and restore binnings that are destroyed by "force_uniform" to avoid
    # breaking following tests
    orig_binnings = binning.BINNINGS.copy()
    plot_calib_distributions.plot_calib_distributions(config)
    binning.BINNINGS = orig_binnings

    plot_path = tmp_path / "DLLK.png"
    assert plot_path.stat().st_size > 2e4

    with open(tmp_path / "plot_calib_distributions.pkl", "rb") as f:
        saved_hist = pickle.load(f)
    assert saved_hist.metadata == {"name": "DLLK"}
    assert saved_hist.values().sum() == pytest.approx(99626.4631959981)
    assert saved_hist.variances().sum() == pytest.approx(128093.73652275163)


@pytest.mark.xrootd
@pytest.mark.slow
def test_create_plot_histograms():
    config = {
        "particle": "K",
        "bin_vars": ["DLLK"],
        "bins": 50,
        "force_uniform": False,
        "cuts": None,
    }
    calib_sample = {
        "files": [
            "root://eoslhcb.cern.ch//eos/lhcb/grid/prod/lhcb/LHCb/Collision15/PIDCALIB.ROOT/00064787/0000/00064787_00000037_1.pidcalib.root",  # noqa
        ]
    }
    tree_paths = ["DSt_KPTuple/DecayTree"]
    branch_names = {"sWeight": "probe_sWeight", "DLLK": "probe_PIDK"}

    hists = plot_calib_distributions.create_plot_histograms(
        config, calib_sample, tree_paths, branch_names
    )

    assert hists[calib_sample["files"][0]]["DLLK"].sum().value == pytest.approx(  # type: ignore # noqa
        6682.033686340834
    )


def test_save_plots(tmp_path):
    hists = {}
    data = np.random.normal(3.5, 2.5, size=100)
    hist = bh.Histogram(bh.axis.Regular(40, -2, 10))
    hist.fill(data)
    hists["test_var"] = hist
    plot_calib_distributions.save_plots(hists, tmp_path, "png")
    plot_path = tmp_path / "test_var.png"
    assert plot_path.stat().st_size > 2e4


def test_save_histograms(tmp_path):
    hists = {}
    data = np.random.normal(3.5, 2.5, size=100)
    hist = bh.Histogram(bh.axis.Regular(40, -2, 10))
    hist.fill(data)
    hists["test_var"] = hist
    plot_calib_distributions.save_histograms(hists, tmp_path)
    with open(tmp_path / "plot_calib_distributions.pkl", "rb") as f:
        saved_hist = pickle.load(f)
    assert saved_hist == hist


def test_decode_arguments():
    assert (
        plot_calib_distributions.decode_arguments(
            [
                "--sample=Turbo18",
                "--magnet=up",
                "--particle=Pi",
                "--bin-var=DLLK",
                "--bin-var=P",
                "--output-dir=pidcalib_output",
            ]
        ).magnet
        == "up"
    )

    with pytest.raises(SystemExit):
        plot_calib_distributions.decode_arguments(
            [
                "--sample=Turbo18",
                "--magnet=up",
                "--bin-var=DLLK",
                "--bin-var=P",
                "--output-dir=pidcalib_output",
            ]
        )



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

"""Convert pickled PIDCalib2 histograms to TH*D & save them in a ROOT file.

Only 1D, 2D, and 3D histograms are supported by ROOT. Attempting to convert
higher-dimensional histograms will result in an exception.
"""

import argparse
import itertools
import logging
import math
import pathlib
import pickle
import sys
from typing import Dict, List, Optional, Union

import boost_histogram as bh
import logzero
import ROOT
from logzero import logger as log

from pidcalib2 import argparse_tools

try:
    from pidcalib2.version import version as VERSION  # type: ignore
except ImportError:
    VERSION = "N/A"


def decode_arguments(args: List[str]) -> argparse.Namespace:
    """Decode CLI arguments."""
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        formatter_class=argparse_tools.RawDefaultsFormatter,
        description=(
            "This tool converts pickled PIDCalib2 histograms to TH*D and saves "
            "them in a ROOT file. It can be used on histograms produced by "
            "make_eff_hists or plot_calib_distributions. Note that ROOT supports "
            "only 1-, 2-, and 3-dimensional histograms; attempting to convert "
            "higher-dimensional histograms will fail."
        ),
    )
    parser.add_argument(
        "input",
        help="path to the input pickle file",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="(debug) increase verbosity",
    )
    parser.add_argument("-V", "--version", action="version", version=VERSION)
    return parser.parse_args(args)


def convert_to_root_histo(
    name: str, bh_histo: bh.Histogram, bh_error_histo: Optional[bh.Histogram] = None
) -> Union[ROOT.TH1D, ROOT.TH2D, ROOT.TH3D]:
    """Convert boost_histogram histogram to a ROOT histogram.

    Only 1D, 2D, and 3D histograms are supported by ROOT. Attempting to convert
    higher-dimensional histograms will result in an exception.

    Furthermore, the boost histogram must have a storage type that stores
    variance, e.g., Weight.

    Args:
        name: Name of the new ROOT histogram.
        bh_histo: The histogram to convert.

    Returns:
        The converted ROOT histogram. Type depends on dimensionality.
    """
    histo = None
    if len(bh_histo.axes) == 1:
        histo = ROOT.TH1D(name, name, 3, 0, 1)
        histo.SetBins(bh_histo.axes[0].size, bh_histo.axes[0].edges)
        histo.GetXaxis().SetTitle(bh_histo.axes[0].metadata["name"])
    elif len(bh_histo.axes) == 2:
        histo = ROOT.TH2D(name, name, 3, 0, 1, 3, 0, 1)
        histo.SetBins(
            bh_histo.axes[0].size,
            bh_histo.axes[0].edges,
            bh_histo.axes[1].size,
            bh_histo.axes[1].edges,
        )
        histo.GetXaxis().SetTitle(bh_histo.axes[0].metadata["name"])
        histo.GetYaxis().SetTitle(bh_histo.axes[1].metadata["name"])
    elif len(bh_histo.axes) == 3:
        histo = ROOT.TH3D(name, name, 3, 0, 1, 3, 0, 1, 3, 0, 1)
        histo.SetBins(
            bh_histo.axes[0].size,
            bh_histo.axes[0].edges,
            bh_histo.axes[1].size,
            bh_histo.axes[1].edges,
            bh_histo.axes[2].size,
            bh_histo.axes[2].edges,
        )
        histo.GetXaxis().SetTitle(bh_histo.axes[0].metadata["name"])
        histo.GetYaxis().SetTitle(bh_histo.axes[1].metadata["name"])
        histo.GetZaxis().SetTitle(bh_histo.axes[2].metadata["name"])
    else:
        raise TypeError(f"{len(bh_histo.axes)}D histograms not supported by ROOT")

    indices_ranges = [list(range(n)) for n in bh_histo.axes.size]
    for indices_tuple in itertools.product(*indices_ranges):
        root_indices = [index + 1 for index in indices_tuple]
        histo.SetBinContent(
            histo.GetBin(*root_indices), bh_histo[indices_tuple].value  # type: ignore
        )
        histo.SetBinError(
            histo.GetBin(*root_indices), math.sqrt(bh_histo[indices_tuple].variance)  # type: ignore # noqa
        )

    return histo


def convert_pklfile_to_rootfile(path: str) -> None:
    """Convert pickled PIDCalib2 histograms to TH*D & save them in a ROOT file.

    The resulting ROOT file has the same name (and path) as the input file, with
    the suffix changed from .pkl to .root.

    Args:
        path: Path of the Pickle file to convert.
    """
    pkl_path = pathlib.Path(path)
    histos: Dict[str, bh.Histogram] = {}
    with open(pkl_path, "rb") as f:
        while True:
            try:
                histo = pickle.load(f)
            except EOFError:
                if not histos:
                    log.error(f"No objects found in '{pkl_path}'")
                    raise
                break

            try:
                histo.metadata["name"]
            except TypeError:
                log.error(
                    f"It seems '{pkl_path}' is not a PIDCalib2 pickle file or "
                    "was made using a version older than 1.1.0"
                )
                raise

            log.debug(f"Unpickling histogram '{histo.metadata['name']}'")
            assert isinstance(histo, bh.Histogram)
            histos[histo.metadata["name"]] = histo

        root_path = pkl_path.with_suffix(".root")
        root_file = ROOT.TFile(str(root_path), "RECREATE")

        for histo in histos.values():
            name = histo.metadata["name"]
            log.info(f"Converting histogram '{histo.metadata['name']}'")
            root_histo = convert_to_root_histo(name, histos[name])
            root_histo.Write()

        log.info(f"Writing ROOT histograms to '{root_path}'")
        root_file.Close()


def main():  # sourcery skip: docstrings-for-functions, require-return-annotation
    config = decode_arguments(sys.argv[1:])
    if config.verbose:
        logzero.loglevel(logging.DEBUG)
    else:
        logzero.loglevel(logging.INFO)
    convert_pklfile_to_rootfile(sys.argv[1])


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import os
import pickle
from pathlib import Path

import pandas as pd
import pytest

from pidcalib2 import utils


@pytest.fixture
def test_path():
    return Path(os.path.dirname(os.path.abspath(__file__)))


def test_make_hist(test_path):
    df = pd.read_csv(str(test_path / "test_data/cal_test_data.csv"), index_col=0)
    hist = utils.make_hist(df, "Pi", ["P"])
    assert hist.size == 20
    assert hist.sum().value == pytest.approx(71.55106080517815)  # type: ignore
    assert hist[3].value == pytest.approx(13.581349537355582)  # type: ignore


def test_create_eff_histograms(test_path):
    df = pd.read_csv(str(test_path / "test_data/cal_test_data.csv"), index_col=0)

    particle = "Pi"
    pid_cut = "DLLK>4"
    bin_vars = ["P"]

    hists = {"total": utils.make_hist(df, particle, bin_vars)}
    df_passing = df.query(pid_cut)
    hists[f"passing_{pid_cut}"] = utils.make_hist(df_passing, particle, bin_vars)

    eff_hists = utils.create_eff_histograms(hists)
    assert eff_hists["eff_DLLK>4"].metadata == {"name": "eff_DLLK>4"}
    assert eff_hists["eff_DLLK>4"].sum().value == pytest.approx(0.18751578358705173)  # type: ignore # noqa
    assert eff_hists["eff_DLLK>4"].size == 20


def test_get_per_event_effs(test_path):
    df_ref = pd.read_csv(str(test_path / "test_data/ref_test_data.csv"), index_col=0)
    prefixes = ["Bach"]
    bin_vars = {"P": "P", "ETA": "ETA", "nTracks": "nTracks"}
    hists = {"Bach": {}}
    with open(
        test_path / "test_data/effhists-Turbo18-up-K-DLLK>4-P.ETA.nTracks.pkl", "rb"
    ) as f:
        hists["Bach"]["eff"] = pickle.load(f)
        hists["Bach"]["passing"] = pickle.load(f)
        hists["Bach"]["total"] = pickle.load(f)
    df_ref = utils.add_bin_indices(df_ref, prefixes, bin_vars, hists)
    # with pytest.warns(RuntimeWarning):
    df_ref = utils.add_efficiencies(df_ref, prefixes, hists)
    assert df_ref.PIDCalibEff.mean() == pytest.approx(0.879822126172073)
    assert df_ref.PIDCalibErr.mean() == pytest.approx(0.0131223551397001)


def test_get_multiparticle_per_event_effs(test_path):
    df_ref = pd.read_csv(str(test_path / "test_data/ref_test_data.csv"), index_col=0)
    prefixes = ["h1", "h2"]
    bin_vars = {"P": "P"}
    hists = {"h1": {}}
    with open(test_path / "test_data/effhists-Turbo18-up-K-DLLK>0-P.pkl", "rb") as f:
        _extracted_from_test_get_multiparticle_per_event_effs_8(hists, "h1", f)
    hists["h2"] = {}
    with open(test_path / "test_data/effhists-Turbo18-up-Pi-DLLK<0-P.pkl", "rb") as f:
        _extracted_from_test_get_multiparticle_per_event_effs_8(hists, "h2", f)
    df_ref = utils.add_bin_indices(df_ref, prefixes, bin_vars, hists)
    df_ref = utils.add_efficiencies(df_ref, prefixes, hists)
    assert df_ref.PIDCalibEff.mean() == pytest.approx(0.928205513381871)
    assert df_ref.PIDCalibErr.mean() == pytest.approx(0.004805856952295592)


def _extracted_from_test_get_multiparticle_per_event_effs_8(hists, arg1, f):
    hists[arg1]["eff"] = pickle.load(f)
    hists[arg1]["passing"] = pickle.load(f)
    hists[arg1]["total"] = pickle.load(f)


def test_add_bin_indices(test_path):
    df_ref = pd.read_csv(str(test_path / "test_data/ref_test_data.csv"), index_col=0)
    prefixes = ["Bach"]
    bin_vars = {"P": "P", "ETA": "ETA", "nTracks": "nTracks"}
    hists = {"Bach": {}}
    with open(
        test_path / "test_data/effhists-Turbo18-up-K-DLLK>4-P.ETA.nTracks.pkl", "rb"
    ) as f:
        hists["Bach"]["eff"] = pickle.load(f)
        hists["Bach"]["passing"] = pickle.load(f)
        hists["Bach"]["total"] = pickle.load(f)
    df_ref = utils.add_bin_indices(df_ref, prefixes, bin_vars, hists)
    assert df_ref["Bach_P_PIDCalibBin"].sum() == 623
    assert df_ref["Bach_ETA_PIDCalibBin"].sum() == 120
    assert df_ref["Bach_PIDCalibBin"].sum() == 10438



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

"""This module handles reading and writing of PID data.

It reads in calibration data locations from a JSON file. It also infers relevant
branch names from the PID cuts and binning variables. Lastly, it infers the PID
histogram file names from the requested configuration.
"""

import collections
import json
import os
import pickle
import re
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import boost_histogram as bh
import pandas as pd
import uproot
from logzero import logger as log

from pidcalib2 import utils
from pidcalib2.aliases import ALIASES
from pidcalib2.samples import SIMPLE_SAMPLES, TUPLE_NAMES


def is_simple(sample: str) -> bool:
    """Return whether a sample has a simple directory structure.

    All Run 1 samples and Run 2 Electron samples have a single DecayTree inside.
    Standard Run 2 samples have multiple directories, each of which has a
    DecayTree inside. This function checks if the sample is in the list of
    simple samples.

    Args:
        sample: Sample name, e.g., Turbo15.
    """
    return sample in SIMPLE_SAMPLES


def get_relevant_branch_names(
    pid_cuts: List[str], bin_vars: List[str], cuts: Optional[List[str]] = None
) -> Dict[str, str]:
    """Return a dict of branch names relevant to the cuts and binning vars.

    The key is user-level variable name/alias. The value is the actual branch
    name.

    Args:
        pid_cuts: Simplified user-level cut list, e.g., ["DLLK < 4"].
        bin_vars: Variables used for the binning.
        cuts: Arbitrary cut list, e.g., ["Dst_IPCHI2 < 10.0"].
    """
    branch_names = {"sWeight": "probe_sWeight"}

    whitespace = re.compile(r"\s+")
    for pid_cut in pid_cuts:
        pid_cut = re.sub(whitespace, "", pid_cut)
        pid_cut_vars = utils.extract_variable_names(pid_cut)

        for pid_cut_var in pid_cut_vars:
            if pid_cut_var not in ALIASES:
                log.warning(
                    (
                        f"PID cut variable '{pid_cut_var}' is not a known alias, "
                        "using raw variable"
                    )
                )
                branch_names[pid_cut_var] = pid_cut_var
            else:
                branch_names[pid_cut_var] = ALIASES[pid_cut_var]

    for bin_var in bin_vars:
        if bin_var not in ALIASES:
            log.warning(
                f"'Binning variable {bin_var}' is not a known alias, using raw variable"
            )
            branch_names[bin_var] = bin_var
        else:
            branch_names[bin_var] = ALIASES[bin_var]

    # Add vars in the arbitrary cuts
    if cuts:
        for cut in cuts:
            cut = re.sub(whitespace, "", cut)
            cut_vars = utils.extract_variable_names(cut)
            for cut_var in cut_vars:
                if cut_var not in ALIASES:
                    log.warning(
                        (
                            f"Cut variable '{cut_var}' is not a known alias, "
                            "using raw variable"
                        )
                    )
                    branch_names[cut_var] = cut_var
                else:
                    branch_names[cut_var] = ALIASES[cut_var]

    if len(branch_names) != len(set(branch_names.values())):
        duplicates = [
            item
            for item, count in collections.Counter(branch_names.values()).items()
            if count > 1
        ]
        log.error(
            (
                "You are mixing aliases and raw variable names for the same "
                f"variable(s): {duplicates}"
            )
        )
        raise KeyError
    return branch_names


def get_reference_branch_names(
    ref_pars: Dict[str, List[str]], bin_vars: Dict[str, str]
) -> List[str]:
    """Return a list of relevant branch names in the reference sample.

    Args:
        ref_pars: A dict of {particle branch prefix : [particle type, PID cut]}
    """
    branch_names = []

    for ref_par_name in ref_pars:
        for bin_var, bin_var_branch in bin_vars.items():
            branch_name = get_reference_branch_name(
                ref_par_name, bin_var, bin_var_branch
            )
            # Avoid duplicate entries
            if branch_name not in branch_names:
                branch_names.append(branch_name)
    return branch_names


def get_reference_branch_name(prefix: str, bin_var: str, bin_var_branch: str) -> str:
    """Return a full name of a binning branch in the reference data.

    Args:
        prefix: Branch prefix of the particle in the reference sample.
        bin_var: Variable used for the binning.
        bin_var_branch: Branch name of the variable used for binning.
    """
    # TODO: Review these hardcoded branch names (maybe consolidate them
    # somewhere). Maybe add some checks that the bin_var is known.
    global_branches = ("nTracks", "nTracks_Brunel", "nSPDhits", "nSPDhits_Brunel")
    if bin_var in global_branches:
        return bin_var_branch

    return f"{prefix}_{bin_var_branch}"


def get_calibration_samples(
    samples_file: Optional[str] = None,
) -> Dict[str, Dict[str, Any]]:
    """Return a dictionary of all files for all calibration samples.

    Args:
        samples_file: JSON file with the calibration file lists.
    """
    if samples_file is None:
        current_dir = os.path.dirname(os.path.abspath(__file__))
        samples_file = str(Path(current_dir, "data/samples.json"))

    log.debug(f"Reading file lists from '{samples_file}'")
    with open(samples_file) as f:
        samples = json.load(f)
    return samples


def get_calibration_sample(
    sample_name: str,
    magnet: str,
    particle: str,
    samples_file: str,
    max_files: Optional[int] = None,
) -> Dict[str, Any]:
    """Return a list of calibration files for a particular sample.

    Args:
        sample: Data sample name (Turbo18, etc.)
        magnet: Magnet polarity (up, down)
        particle: Particle type (K, pi, etc.)
        samples_file: File in which to look up the file list.
        max_files: Optional. The maximum number of files to get. Defaults to
            None (= all files).
    """
    samples = get_calibration_samples(samples_file)

    magnet = f"Mag{magnet.capitalize()}"
    sample_full_name = "-".join([sample_name, magnet, particle])

    try:
        sample = samples[sample_full_name]
    except KeyError:
        log.error(
            (
                f"Sample '{sample_full_name}' not found. "
                "Consult 'pidcalib2.make_eff_hists --list configs'."
            )
        )
        raise

    calibration_sample = sample.copy()
    if "link" in sample:
        del calibration_sample["link"]
        link = sample["link"]
        try:
            calibration_sample["files"] = samples[link]["files"]
        except KeyError:
            log.error(f"Linked sample '{link}' not found in {samples_file}")
            raise
        # Copy configuration from the link (tuple_names, cuts, etc.), but only
        # if it doesn't override the upstream configuration
        for key, item in samples[link].items():
            if key == "files":
                continue
            if key not in calibration_sample:
                calibration_sample[key] = item

    if max_files:
        log.warning(
            "You are using the --max-files option; it should be used only for testing"
        )
        calibration_sample["files"] = calibration_sample["files"][:max_files]
    return calibration_sample


def root_to_dataframe(
    path: str, tree_names: List[str], branches: List[str], calibration: bool = False
) -> Union[pd.DataFrame, None]:  # sourcery skip: extract-duplicate-method
    """Return DataFrame with requested branches from tree in ROOT file.

    Args:
        path: Path to the ROOT file; either file system path or URL, e.g.
            root:///eos/lhcb/file.root.
        tree_names: Names of trees inside the ROOT file to read.
        branches: Branches to put in the DataFrame.
        calibration: Whether the file is a calibration sample or a user sample.
    """
    tree = None
    # EOS sometimes fails with a message saying the operation expired. It is
    # intermittent and hard to replicate. See this related issue:
    # https://github.com/scikit-hep/uproot4/issues/351. To avoid PIDCalib2
    # completely failing in these cases, we skip the file with a warning
    # message if this happens.
    try:
        root_file = uproot.open(path)
    except FileNotFoundError as exc:
        if "Server responded with an error: [3010]" in exc.args[0]:
            log.error(
                (
                    "Permission denied; do you have Kerberos credentials? "
                    "Try running 'kinit -f [USERNAME]@CERN.CH'"
                )
            )
        raise
    except OSError as err:
        if "Operation expired" not in err.args[0]:
            raise
        log.error(
            f"Failed to open '{path}' because an XRootD operation expired; skipping"
        )
        print(err)
        return None

    dfs = []
    for tree_name in tree_names:
        try:
            tree = root_file[tree_name]
            dfs.append(tree.arrays(branches, library="pd"))  # type: ignore
        except uproot.exceptions.KeyInFileError as exc:  # type: ignore
            assert tree is not None
            similar_keys = []
            if calibration:
                aliases_in_tree = [
                    alias for alias, var in ALIASES.items() if var in tree.keys()
                ]

                similar_keys += utils.find_similar_strings(
                    exc.key, aliases_in_tree, 0.80
                )
                similar_keys += utils.find_similar_strings(
                    "probe_" + exc.key, tree.keys(), 0.80
                )
            similar_keys += utils.find_similar_strings(exc.key, tree.keys(), 0.80)
            similar_keys += utils.find_similar_strings(
                exc.key.replace("Brunel", ""), tree.keys(), 0.80
            )
            # Remove duplicates while preserving ordering
            similar_keys = list(dict.fromkeys(similar_keys))
            log.error(
                (
                    f"Branch '{exc.key}' not found; similar aliases and/or branches "
                    f"that exist in the tree: {similar_keys}"
                )
            )
            raise
        except OSError as err:
            if "Operation expired" not in err.args[0]:
                raise

            log.error(
                (
                    f"Failed to open '{path}' because an XRootD operation "
                    "expired; skipping"
                )
            )
            print(err)
            return None
    return pd.concat(dfs, ignore_index=True)


def get_tree_paths(
    particle: str,
    sample: str,
    override_tuple_names: Optional[List[str]] = None,
) -> List[str]:
    """Return a list of internal ROOT paths to relevant trees in the files

    Args:
        particle: Particle type (K, pi, etc.)
        sample: Data sample name (Turbo18, etc.)
        override_tuple_names: Tree names to use instead of the default
    """
    tree_paths = []
    if is_simple(sample):
        # Run 1 (and Run 2 Electron) files have a simple structure with a single
        # tree
        tree_paths.append("DecayTree")
    elif override_tuple_names:
        log.debug("Tree paths overriden by tuple_names")
        tree_paths.extend(
            f"{tuple_name}/DecayTree" for tuple_name in override_tuple_names
        )

    else:
        tree_paths.extend(
            f"{tuple_name}/DecayTree" for tuple_name in TUPLE_NAMES[particle]
        )

    return tree_paths


def dataframe_from_local_file(path: str, branch_names: List[str]) -> pd.DataFrame:
    """Return a dataframe read from a local file (instead of EOS).

    Args:
        path: Path to the local file.
        branch_names: Columns to read from the DataFrame.
    """
    if path.endswith(".pkl"):
        df = pd.read_pickle(path)
    elif path.endswith(".csv"):
        df = pd.read_csv(path, index_col=0)
    else:
        log.error(
            (
                f"Local dataframe file '{path}' "
                f"has an unknown suffix (csv and pkl supported)"
            )
        )
        raise TypeError("Only csv and pkl files supported")
    log.info(f"Read {path} with a total of {len(df.index)} events")

    try:
        df = df[branch_names]
    except KeyError:
        log.error("The requested branches are missing from the local file")
        raise

    return df


def get_calib_hists(
    hist_dir: str,
    sample: str,
    magnet: str,
    ref_pars: Dict[str, List[str]],
    bin_vars: Dict[str, str],
) -> Dict[str, Dict[str, bh.Histogram]]:
    """Get calibration efficiency histograms from all necessary files.

    Args:
        hist_dir: Directory where to look for the required files.
        sample: Data sample name (Turbo18, etc.).
        magnet: Magnet polarity (up, down).
        ref_pars: Reference particle prefixes with a particle type and PID cut.
        bin_vars: Binning variables ({standard name: reference sample branch name}).

    Returns:
        Dictionary with an efficiency histogram for each reference particle.
        The reference particle prefixes are the dictionary keys.
    """
    hists: Dict[str, Dict[str, bh.Histogram]] = {}
    for ref_par in ref_pars:
        particle = ref_pars[ref_par][0]

        pid_cut = ref_pars[ref_par][1]
        whitespace = re.compile(r"\s+")
        pid_cut = re.sub(whitespace, "", pid_cut)

        calib_name = Path(
            hist_dir,
            utils.create_hist_filename(
                sample, magnet, particle, pid_cut, list(bin_vars)
            ),
        )

        log.debug(f"Loading efficiency histograms from '{calib_name}'")

        hists[ref_par] = {}
        try:
            with open(calib_name, "rb") as f:
                hists[ref_par]["eff"] = pickle.load(f)
                hists[ref_par]["passing"] = pickle.load(f)
                hists[ref_par]["total"] = pickle.load(f)
        except FileNotFoundError:
            log.error(
                (
                    "Efficiency histogram file not found. You must first "
                    "run make_eff_hists with matching parameters to create the "
                    "efficiency histograms."
                )
            )
            raise
    return hists


def save_dataframe_as_root(
    df: pd.DataFrame, name: str, filename: str, columns: Optional[List[str]] = None
) -> None:
    """Save a DataFrame as a TTree in a ROOT file.

    NaN entries are changed to -999 because ROOT TTrees don't support NaNs.

    Args:
        df: DataFrame to be saved.
        name: Name of the new TTree.
        filename: Name of the file to which to save the TTree.
        columns: Optional. Names of the columns which are to be saved. If
            'None', all the columns will be saved.
    """
    df_wo_nan = df.fillna(-999)
    if columns is None:
        columns = list(df_wo_nan.keys())
    branches_w_types = {branch: df_wo_nan[branch].dtype for branch in columns}
    with uproot.recreate(filename) as f:
        log.debug(f"Creating a TTree with the following branches: {branches_w_types}")
        f.mktree(name, branches_w_types)
        branch = {
            branch_name: df_wo_nan[branch_name] for branch_name in branches_w_types
        }
        f[name].extend(branch)
    log.info(f"Efficiency tree saved to {filename}")



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import math
import os
import pickle
import shutil
from pathlib import Path

import pytest
import uproot

try:
    from pidcalib2 import pklhisto2root
except ModuleNotFoundError:
    pytest.skip("PyROOT not available", allow_module_level=True)


@pytest.fixture
def test_path():
    return Path(os.path.dirname(os.path.abspath(__file__)))


@pytest.mark.pyroot
def test_pklhisto2root_old(test_path, tmp_path):
    shutil.copy(
        test_path / "test_data/effhists-Turbo18-up-K-DLLK>0-P-pre1.1.0.pkl",
        tmp_path / "effhists-Turbo18-up-K-DLLK>0-P-pre1.1.0.pkl",
    )
    pkl_path = tmp_path / "effhists-Turbo18-up-K-DLLK>0-P-pre1.1.0.pkl"
    # Test that using old pickle files without metadata raises an error
    with pytest.raises(TypeError):
        pklhisto2root.convert_pklfile_to_rootfile(pkl_path)


@pytest.mark.pyroot
def test_pklhisto2root(test_path, tmp_path):
    shutil.copy(
        test_path / "test_data/effhists-Turbo18-up-K-DLLK>0-P.pkl",
        tmp_path / "effhists-Turbo18-up-K-DLLK>0-P.pkl",
    )
    pkl_path = tmp_path / "effhists-Turbo18-up-K-DLLK>0-P.pkl"
    root_path = tmp_path / "effhists-Turbo18-up-K-DLLK>0-P.root"
    pklhisto2root.convert_pklfile_to_rootfile(pkl_path)
    names = ["eff_DLLK>0", "passing_DLLK>0", "total"]
    boost_hists = {}
    with open(pkl_path, "rb") as f:
        for name in names:
            boost_hists[name] = pickle.load(f)

    assert root_path.exists()
    root_file = uproot.open(root_path)

    for name in names:
        assert root_file[name].values()[0] == pytest.approx(
            boost_hists[name].values()[0]
        )
        assert root_file[name].errors()[0] == pytest.approx(  # type: ignore
            math.sqrt(boost_hists[name].variances()[0])
        )



"""A module for printing simple Markdown tables.

Each column can have a different numeric precision, any cell can have a string
instead of a number.

Usage:
    table = MarkdownTable(["Column1", "Column2"])
    table.add_row(["Row1", 3.14])
    table.set_precision([0, 1])
    table.print()
"""

from typing import List, Union


class MarkdownTable:
    """Class representing a Markdown table with support for pretty printing."""

    def __init__(self, header):
        """Initialize table with supplied column titles.

        Args:
            header: A list containing column titles.
        """
        assert type(header) in [list, tuple]
        self.header = header
        self.rows = []
        self.precision = [3 for _ in header]

    def set_precision(self, precisions: List[int]) -> None:
        """Set columns' decimal precisions.

        Applies only to numeric cells.

        Args:
            precisions: A list of integers representing number of decimals in
                each column.
        """
        assert len(precisions) == len(self.header)
        assert type(precisions) in [list, tuple]
        self.precision = precisions

    def add_row(self, row: List[float]) -> None:
        """Add a single row.

        The row must be the same length as the header used to initialize the
        table.
        """
        assert len(self.header) == len(row)
        self.rows.append(row)

    # def add_midline(self):
    #     self.rows.append(["" for _ in range(len(header))])

    def get_length_of_number(self, number: float, precision: int) -> int:
        """Return the length of a number written with a specified precision."""
        number_format = "{:." + str(precision) + "f}"
        return len(number_format.format(number))

    def get_column_widths(self) -> List[int]:
        """Return the maximum width of content in all columns."""
        columns = []
        for col in range(len(self.header)):
            cells = []
            for row in self.rows:
                if type(row[col]) is str:
                    cells.append(len(row[col]))
                else:
                    cells.append(
                        self.get_length_of_number(row[col], self.precision[col])
                    )
            cells.append(len(self.header[col]))
            columns.append(max(cells))
        return columns

    def print(self) -> None:
        """Print the table."""
        self.columns_widths = self.get_column_widths()
        self.print_row(self.header)
        self.print_separator()
        for row in self.rows:
            self.print_row(row)

    def get_format_string(self, width: int, precision: Union[int, None] = None) -> str:
        """Get a formatting string.

        You normally write these yourself, but here we want all the rows in a
        column to have the same width, regardless of content.

        Args:
            width: Total requested width of the string.
            precision: Optional precision for numeric cells.
        """
        format_string = "{:" + str(width)
        if precision is not None:
            format_string += f".{str(precision)}" + "f} | "
        else:
            format_string += "} | "
        return format_string

    def print_row(self, row: List[Union[str, float]]) -> None:
        """Print a single row."""
        row_text = ""
        for col, cell in enumerate(row):
            format_string = self.get_format_string(
                self.columns_widths[col],
                (None if type(cell) is str else self.precision[col]),
            )
            row_text += format_string.format(cell)

        print(row_text.rstrip(" |"))

    def print_separator(self) -> None:
        """Print a separator row (midrule)."""
        print("|".rjust(self.columns_widths[0] + 2, "-"), end="")
        for col, _ in enumerate(self.header[1:-1]):
            print("|".rjust(self.columns_widths[col + 1] + 3, "-"), end="")
        print("".ljust(self.columns_widths[-1] + 1, "-"))
